



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
meson -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
meson -v
IF %ERRORLEVEL% NEQ 0 exit /B 1
meson --internal delsuffix . nomatchqqq
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
